////////////////////////////////////////////////////////////////////////////////
///
///   Copyright (C) 2006-2008   Brian C. Becker             www.BrianCBecker.com
///                             Enrique G. Ortiz           www.EnriqueGOrtiz.com
///   License: LGPL             RI @ CMU                          www.ri.cmu.edu
///
///   Description: Facebook Friend Picture Downloader
///   --------------------------------------------------------------------------
///	  Logs into your Facebook account and automatically downloads all your 
///   friend's pictures. This may take a while as your friends may have uploaded 
///   a bunch of pictures. It is not unusual to have >2GB of picture files if 
///   you have several hundred friends. In addition to downloading friend's 
///   pictures, this app downloads tags associated with the pictures. Photos 
///   without tags are not downloaded as you can't really extract any meaningful
///   information.
/// 
///   In order for this to work you need to put your Application and Secret Key
///   from Facebook on the 2nd and 3rd lines (respectively) of the "facebook
///   api key.txt" Also, in order to download somebody's pictures, they need to
///   have added the application to their Facebook page or else be listed as a
///   developer (usually an easier way to do things).
/// 
///   See author's website for more information and contact/bug reports.
///
////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Collections.ObjectModel;
using Facebook.Entity;
using System.IO;
using AT.MIN;
using System.Drawing.Imaging;

namespace VS_Facebook
{
    public partial class frmFacebook : Form
    {
        Collection<String> ids;                   ///< Collection of friend IDs
        Queue<User> users = new Queue<User>();    ///< Queue of Facebook friends's to process
        Collection<User> infos;                   ///< Information about friends
        Collection<Photo> photos;                 ///< Pictures a friend has uploaded
        Collection<PhotoTag> tags;                ///< Tags for an image (where a friend is)
        Boolean fastDone = false;                 ///< Write out a done file for each user
        Boolean stopping = false;                 ///< Has the app been closed?

        /// The directory that all the information gets saved to
        String dir = "";

        public frmFacebook()
        {
            InitializeComponent();
        }

        /// <summary>
        /// Loads the Facebook API/secret key from file and prompts user to log in
        /// NOTE: User MUST have added the Facebook app to their profile (or be a developer)
        /// </summary>
        private void loadFacebook(object sender, EventArgs e)
        {
            // Prompt an immediate login, getting application key from file
            string filename = "facebook api key.txt";
            try
            {
                StreamReader SR;
                SR = File.OpenText(filename);
                SR.ReadLine(); // Ignore first line, it's a comment
                fbs.ApplicationKey = SR.ReadLine(); // Application Key for Facebook app
                fbs.Secret = SR.ReadLine(); // Secret Key for Facebook app
                SR.Close();
            }
            catch
            {
                MessageBox.Show("Error: File \"" + filename + "\" must contain your Facebook Application & Secret Key, on line 2 & 3 respectively.");
                Application.Exit();
            }
            fbs.ConnectToFacebook();
        }

        /// <summary>
        /// If the user just selected a directory, display on GUI
        /// </summary>
        private void btnBrowse_Click(object sender, EventArgs e)
        {
            if (fbdBrowse.ShowDialog() == DialogResult.OK)
            {
                txtFolder.Text = fbdBrowse.SelectedPath;
            }
        }

        /// <summary>
        /// 3rd party function to grab encoders (forgot where I got it from)
        /// </summary>
        /// <param name="mimeType">MIME type, used for jpeg</param>
        private static ImageCodecInfo GetEncoderInfo(String mimeType)
        {
            int j;
            ImageCodecInfo[] encoders;
            encoders = ImageCodecInfo.GetImageEncoders();
            for (j = 0; j < encoders.Length; ++j)
            {
                if (encoders[j].MimeType == mimeType)
                    return encoders[j];
            }
            return null;
        }

        /// <summary>
        /// Saves an image to disk with specified filename and compression in JPEG format
        /// </summary>
        /// <param name="image">Image to save</param>
        /// <param name="szFileName">Filename to save as</param>
        /// <param name="lCompression">Percentage JPEG compression to save with</param>
        private void SaveJPGWithCompressionSetting(Image image, string szFileName, long lCompression)
        {
            EncoderParameters eps = new EncoderParameters(1);
            eps.Param[0] = new EncoderParameter(System.Drawing.Imaging.Encoder.Quality, lCompression);
            ImageCodecInfo ici = GetEncoderInfo("image/jpeg");
            image.Save(szFileName, ici, eps);
        }

        /// <summary>
        /// Saves an image to disk with specified filename and compression in PNG format
        /// </summary>
        /// <param name="image">Image to save</param>
        /// <param name="szFileName">Filename to save as</param>
        /// <param name="lCompression">Percentage PNG compression to save with</param>
        private void SavePNGWithCompressionSetting(Image image, string szFileName, long lCompression)
        {
            EncoderParameters eps = new EncoderParameters(1);
            eps.Param[0] = new EncoderParameter(System.Drawing.Imaging.Encoder.Quality, lCompression);
            ImageCodecInfo ici = GetEncoderInfo("image/png");
            image.Save(szFileName, ici, eps);
        }

        /// <summary>
        /// Main loop of the code, downloads friend's pictures and the tags 
        /// associated with the pictures. A separate folder is created for each user's photos
        /// Under each user's folder, the pictures and tags are stored in a folder called raw
        /// The idea being that you will process the pictures and tags and store them elsewhere
        /// 
        /// </summary>
        private void btnStart_Click(object sender, EventArgs e)
        {
            // Make sure the user has selected a valid directory
            if (txtFolder.Text == "" || !Directory.Exists(txtFolder.Text))
            {
                MessageBox.Show("Error: You must select a valid folder to store the data in.");
                return;
            }

            lblCurFriend.Text = "Loading Friends...";
            Application.DoEvents();

            // Get all the main Facebook user's friends
            ids = fbs.GetFriendIds();

            // We want the main Facebook user's photos too
            ids.Insert(0,fbs.UserId);

            // Grab information for all users
            infos = new Collection<User>();
            pgbFriend.Value = 0;
            pgbFriend.Minimum = 0;
            pgbFriend.Maximum = ids.Count;
            btnStart.Enabled = false;
            for (int i = 0; i < ids.Count; i++)
            {
                pgbFriend.Value++;
                Collection<User> u = null;
                Collection<string> tmp = new Collection<string>();
                tmp.Add(ids[i]);
                int ctr = 0;
                do
                {
                    ctr++;
                    try
                    {
                        u = fbs.GetUserInfo(tmp);
                    }
                    catch
                    {
                        lblCurFriend.Text = "Error!";
                    }
                } while (ctr < 10 && (u == null || u.Count == 0));

                if (u != null && u.Count > 0)
                {
                    lbxFriends.Items.Add(u[0].Name + " (" + u[0].UserId + ")");
                    infos.Add(u[0]);
                    users.Enqueue(u[0]);
                }

                Application.DoEvents();
            }
            btnStart.Enabled = true;
            lblCurFriend.Text = "";

            // Display a progress bar for the friends we want to process
            pgbOverall.Minimum = 0;
            pgbOverall.Maximum = infos.Count; // When we finish with friends we are done
            pgbOverall.Value = 0;

            // Disable start button and grab the directory
            btnStart.Enabled = false;
            dir = txtFolder.Text;

            // Check if dir has been updated
            if (dir[dir.Length - 1] != '\\' && dir[dir.Length - 1] != '/')
            {
                dir += '/';
                txtFolder.Text = dir;
            }

            // Loop through all the users to try to find if they 
            while (users.Count > 0 && !stopping)
            {
                // Grab user & user ID
                User user = users.Dequeue();
                String friend = user.UserId;

                // Friend directory
                String friendDir = user.Name.Replace("(", "") + " (" + friend + ")";

                // We put this inside a try...catch block because facebook api can time out, 
                // in which case we want to try to get the rest of the pictures/tags from
                // this user later
                try
                {
                    // When we are done with this friend, we can set a file
                    String doneFile = friend + ".txt";

                    // Create friend directory if it doesn't exist
                    if (!Directory.Exists(dir + friendDir))
                        Directory.CreateDirectory(dir + friendDir);

                    // Raw data (pictures + tags)
                    friendDir = friendDir + "/raw/";
                    if (!Directory.Exists(dir + friendDir))
                        Directory.CreateDirectory(dir + friendDir);

                    //if (File.Exists(dir + doneFile))
                    //    continue;

                    //lbxFriends.SelectedIndex = i;

                    // Update which friend we are processing
                    lblCurFriend.Text = user.Name;
                    Application.DoEvents();

                    // Get photos for user and setup progress bars
                    photos = fbs.GetPhotos(user);
                    pgbFriend.Minimum = 0;
                    pgbFriend.Maximum = photos.Count;
                    pgbFriend.Value = 0;
                    String allTags = "";

                    // Grab all the photos
                    foreach (Photo photo in photos)
                    {
                        if (stopping)
                            break;

                        // Names for photos and tags
                        String photoname = dir + friendDir + friend + "." + photo.PhotoId + ".jpg";
                        String tagsname = dir + friendDir + friend + "." + photo.PhotoId + ".txt";

                        // If the tags exist, we have the photo
                        if (File.Exists(tagsname))
                            continue;

                        // Grab photo
                        Collection<String> singlePhoto = new Collection<String>();
                        singlePhoto.Add(" " + photo.PhotoId);
                        tags = fbs.GetTags(singlePhoto);

                        // If nobody is tagged, don't bother downloading it
                        if (tags.Count > 0)
                        {
                            // Save as a jpeg with 90% compression
                            lblImgName.Text = friend + "." + photo.PhotoId + ".jpg";
                            Image image = new Bitmap(photo.PictureBig);
                            SaveJPGWithCompressionSetting(image, photoname, 90); 
                            //image.Save(photoname, System.Drawing.Imaging.ImageFormat.Jpeg);

                            // Draw image on the screen
                            Graphics g = Graphics.FromImage(image);
                            float width, height, x, y, size;
                            width = image.Width;
                            height = image.Height;
                            size = 25;
                            Pen p = new Pen(Color.Red, 5);

                            // Draw tags on the screen
                            allTags = "";
                            foreach (PhotoTag tag in tags)
                            {
                                x = (float)tag.XCoord;
                                y = (float)tag.YCoord;
                                allTags += friend + "." + tag.PhotoId + "." + tag.SubjectUserId + " " + tag.XCoord + " " + tag.YCoord + "\n";

                                g.DrawLine(p, x * width / 100 - size, y * height / 100, x * width / 100 + size, y * height / 100);
                                g.DrawLine(p, x * width / 100, y * height / 100 - size, x * width / 100, y * height / 100 + size);
                            }

                            // Write tags to file
                            StreamWriter sw = new StreamWriter(tagsname, true);
                            sw.Write(allTags);
                            sw.Flush();
                            sw.Close();

                            // Clean up
                            g.Dispose();
                            p.Dispose();
                            pbxImg.Image = image;

                            pgbFriend.Value++;
                        }
                        Application.DoEvents();
                    }

                    // Create the done file to signify we are done with this user
                    if (fastDone)
                    {
                        StreamWriter done = new StreamWriter(dir + doneFile, false);
                        done.Close();
                    }

                    pgbOverall.Value++;

                    lbxFriends.Items.Remove(user.Name + " (" + user.UserId + ")");
                }
                catch (Exception exception)
                {
                    // If something crashed (facebook timeout or something like that),
                    // throw the facebook user back on the queue and try again later
                    System.Console.WriteLine(exception.Message);
                    users.Enqueue(user);
                    //pgbOverall.Maximum++;
                }

                Application.DoEvents();
            }

            btnStart.Enabled = true;
            pgbOverall.Value = 0;
            pgbFriend.Value = 0;
            pbxImg.Image = new Bitmap(100, 100);
            lblCurFriend.Text = "Done";
            lblImgName.Text = "";
        }

        /// <summary>
        /// If the user has closed the form in the middle of downloading, set a 
        /// flag so that the downloading process can quit and allow the app to close
        /// </summary>
        private void frmFacebook_FormClosing(object sender, FormClosingEventArgs e)
        {
            stopping = true;
        }

        private void llbAbout_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            System.Diagnostics.Process.Start("http://www.BrianCBecker.com");
        }
    }
}